#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>

from . import bounding
from . import draw
from . import errors
from . import projection
from . import svg
from . import style
from . import svgis
from . import transform
from .svgis import map, SVGIS

__version__ = '0.4.1'

__all__ = [
    'bounding',
    'draw',
    'errors',
    'projection',
    'style',
    'svg',
    'svgis',
    'transform',
]



# -*- coding: utf-8 -*-

"""Draw SVG maps"""

# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
import os.path
from collections import Iterable
from functools import partial
import logging
import fiona
import fiona.transform
import fionautil.scale
from six import string_types
from . import bounding, draw, errors, projection, svg, transform, utils
from . import style as _style


STYLE = ('polyline,line,rect,path,polygon,.polygon{'
         'fill:none;'
         'stroke:#000;'
         'stroke-width:1px;'
         'stroke-linejoin:round;'
         '}')


def map(layers, bounds=None, scale=None, **kwargs):
    '''
    Draw a geodata layer to SVG. This is shorthand for creating a :class:`SVGIS` instance
    and immediately runnning :class:`SVGIS.compose`.

    Args:
        layers (sequence): Input geodata files.
        output (path): Output file name
        bounds (sequence): (minx, miny, maxx, maxy)
        scale (int): Map scale. Larger numbers -> smaller maps
        padding (int): Pad around bounds by this much. In projection units.
        crs (string): EPSG code, PROJ.4 string, or file containing a PROJ.4 string
        clip (bool): If true, clip features output to bounds.
        style (Sequence): Path to a css file or a css string.
        class_fields (Sequence): A comma-separated string or list of class names to
                                 use the SVG drawing.
        id_field (string): Field to use to determine id of each element in the drawing.
        inline (bool): If False, do not move CSS declarations into each element.
        precision (int): Precision for rounding output coordinates.

    Returns:
        String (unicode in Python 2) containing an entire SVG document.
    '''
    scale = (1. / scale) if scale else 1.
    bounds = bounding.check(bounds)

    # Try to read style file(s)
    styles = u''.join(_style.pick(s) for s in kwargs.pop('style', []))

    class_fields = set(a for c in kwargs.pop('class_fields', []) for a in c.split(','))

    drawing = SVGIS(
        layers,
        bounds=bounds,
        scalar=scale,
        crs=kwargs.pop('crs', None),
        style=styles,
        clip=kwargs.pop('clip', True),
        id_field=kwargs.pop('id_field', None),
        class_fields=class_fields,
        simplify=kwargs.pop('simplify', None)
    ).compose(**kwargs)

    return drawing


class SVGIS(object):

    """
    Draw geodata files to SVG.

    Args:
        files (list): A list of files to draw.
        bounds (Sequence): An iterable with four float coordinates in (minx, miny, maxx, maxy) format
        crs (dict): A proj-4 like mapping, or a projection method keyword (file, local, utm).
        style (string): CSS to add to output file
        scalar (int): Map scaling factor (output coordinate are multiplied by this)
        style (str): CSS styles
        padding (number): Buffer each edge by this many map units.
        precision (int): Precision for rounding output coordinates.
        simplify (int): Simplification factor (between 1 and 100).
        id_field (str): Field in data to use for ID'ing elements.
        class_fields (Sequence): Fields in data for added classes to elements.
    """

    # The bounding box in input coordinates.
    _unprojected_bounds = (None,) * 4

    # The bounding box in output coordinates, to be determined as we draw.
    _projected_bounds = (None,) * 4

    _in_crs, _out_crs = None, None

    clipper = None
    simplifier = None

    def __init__(self, files, bounds=None, crs=None, **kwargs):
        self.log = logging.getLogger('svgis')

        if isinstance(files, string_types):
            self.files = [files]
        elif isinstance(files, Iterable):
            self.files = files
        else:
            raise ValueError("'files' must be a file name or list of file names")

        self.log.info('starting SVGIS, files: %s', ', '.join(self.files))

        if bounding.check(bounds):
            self._unprojected_bounds = bounds
        elif bounds:
            self.log.warn("ignoring invalid bounds: %s", bounds)

        # This may return a keyword, which will require more updating.
        # If so, will update when files are open.
        self._out_crs = projection.pick(crs)

        self.scalar = kwargs.pop('scalar', 1) or 1

        self.style = STYLE + (kwargs.pop('style', '') or '')

        self.padding = kwargs.pop('padding', 0) or 0

        self.precision = kwargs.pop('precision', None)

        self.clip = kwargs.pop('clip', True)

        simple = kwargs.pop('simplify', None)

        if simple:
            self.simplifier = transform.simplifier(simple)
            self.log.debug('Simplifying with a factor of %d', simple)

        self.id_field = kwargs.pop('id_field', None)

        self.class_fields = kwargs.pop('class_fields', [])

    def __repr__(self):
        return ('SVGIS(files={0.files}, out_crs={0.out_crs})').format(self)

    @property
    def in_crs(self):
        return self._in_crs

    def set_in_crs(self, crs):
        if not self.in_crs and crs:
            self._in_crs = crs

        if not crs:
            # Assume input CRS is WGS 84
            self._in_crs = utils.DEFAULT_GEOID
            self.log.warn('Found no input coordinate system, '
                          'assuming WGS84 (long/lat) coordinates.')

    @property
    def out_crs(self):
        if self._out_crs in projection.METHODS:
            return None
        else:
            return self._out_crs

    def set_out_crs(self, bounds):
        '''Set the output CRS, if not yet set.'''
        if self.out_crs:
            return

        # Determine projection transformation:
        # either use something passed in, a non latlong layer projection,
        # the local UTM, or customize local TM
        self.log.debug('Picking a projection:')
        self.log.debug('    out crs: %s', self._out_crs)
        self.log.debug('    in crs: %s', projection.fake_to_string(self.in_crs))
        self.log.debug('    bounds: %s', bounds)
        self._out_crs = projection.pick(self._out_crs, bounds, self.in_crs)
        self.log.debug('Set output crs to %s', projection.fake_to_string(self.out_crs))

    @property
    def unprojected_bounds(self):
        '''Returns None if projected bounds aren't set'''
        if not all(self._unprojected_bounds):
            return None
        return self._unprojected_bounds

    @property
    def projected_bounds(self):
        '''Returns None if projected bounds aren't (yet) set'''
        if not all(self._projected_bounds):
            return None
        return self._projected_bounds

    def update_projected_bounds(self, in_crs, out_crs, bounds, padding=None):
        '''
        Extend projected_bounds bbox with self.padding.

        Args:
            in_crs (dict): CRS of bounds.
            out_crs (dict) desired output CRS.
            bounds (tuple): bounding box.

        Returns:
            (tuple) bounding box in out_crs coordinates.
        '''
        # This may happen many times if we were passed bounds, but it's a cheap operation.
        projected = bounding.transform(in_crs, out_crs, bounds)
        self._projected_bounds = bounding.pad(projected, padding or 0)
        return self._projected_bounds

    def _get_clipper(self, layer_bounds, out_bounds, scalar=None):
        '''
        Get a clipping function for the given input crs and bounds.

        Args:
            layer_bounds (tuple): The bounds of the layer.
            out_bounds (tuple): The desired output bounds (in layer coordinates).
            scalar (float): Map scale.

        Returns:
            None if layer_bounds are inside out_bounds or clipping is off.
        '''
        if not self.clip or bounding.covers(out_bounds, layer_bounds):
            return None

        scalar = scalar or self.scalar

        if not self.clipper:
            padded_bounds = bounding.pad(self.projected_bounds, 1000)
            self.clipper = transform.clipper([c * scalar for c in padded_bounds])

        return self.clipper

    def _reprojector(self, in_crs):
        '''Return a reprojection transform from in_crs to self.out_crs.'''
        if self.out_crs != in_crs:
            self.log.info('set up reprojection')
            self.log.debug('input crs: %s', projection.fake_to_string(in_crs))
            return partial(fiona.transform.transform_geom, in_crs, self.out_crs)

        else:
            return None

    def _prepare_layer(self, layer, filename, bounds, scalar, **kwargs):
        '''
        Prepare the keyword args for drawing a layer.

        Args:
            layer (fiona.layer): input layer
            filename (str): Name of file, used for group id attribute.
            bounds (tuple): Bounding box (in layer.crs).
            scalar (int): Map scale
            simplifier (function): simplication function
            id_field (str): Field to use for element id attribute.
            class_fields (list): Fields to use for element class attribute.

        Returns:
            (dict) Arguments for self._feature
        '''
        result = {
            'transforms': [
                self._reprojector(layer.crs),
                partial(fionautil.scale.geometry, factor=scalar),
                # Get clipping function based on a slightly extended version of _projected_bounds.
                self._get_clipper(layer.bounds, bounds, scalar=scalar),
                self.simplifier
            ]
        }

        # Correct for OGR's lack of creativity for GeoJSONs.
        if layer.name == 'OGRGeoJSON':
            result['name'] = os.path.splitext(os.path.basename(filename))[0]
        else:
            result['name'] = layer.name

        # A list of class names to get from layer properties.
        class_fields = kwargs.pop('class_fields', None) or self.class_fields
        result['classes'] = [x for x in class_fields if x in layer.schema['properties']]

        # Remove the id field if it doesn't appear in the properties.
        id_field = kwargs.pop('id_field', self.id_field)

        result['id_field'] = id_field if id_field in layer.schema['properties'].keys() else None

        result.update(kwargs)

        return result

    def compose_file(self, path, unprojected_bounds=None, **kwargs):
        '''
        Draw fiona file to an SVG group.

        Args:
            path (string): path to a fiona-readable file, or an Apache Commons VFS spec for a zip
                           or tar archive, e.g. ``zip://path/to/archive.zip/file.shp``.
            unprojected_bounds (tuple): (minx, maxx, miny, maxy) in the layer's coordinate system.
                                        'None' values are OK. "Unprojected" here refers to
                                        the fact that we haven't transformed these bounds yet.
                                        They may well, in fact, be in a projection.
            padding (int): Number of map units by which to pad output bounds.
            scalar (int): map scale
            class_fields (sequence): Fields to turn in the element classes (default: self.class_fields).
            id_field (string): Field to use as element ID (default: self.id_field).

        Returns:
            dict with keys: members, id, class. This is ready to be passed to svgis.svg.group.
        '''
        padding = kwargs.pop('padding', self.padding)
        kwargs['scalar'] = kwargs.get('scalar', self.scalar)
        unprojected_bounds = unprojected_bounds or self.unprojected_bounds
        vfs = None

        # "Detect Virtual File System" sounds cooler than "Does It Start with zip or tar".
        if path.startswith('zip://') or path.startswith('tar://'):
            suffix = '.zip' if path.startswith('z') else '.gz'
            vfs, path = path.split(suffix, 1)
            vfs = vfs + suffix
            self.log.debug('reading vfs: %s ', vfs)

        with fiona.drivers():
            self.log.debug('opening %s', path)
            with fiona.open(path, vfs=vfs) as layer:
                self.log.info('reading %s', layer.name)

                # Set the input CRS, if not yet set.
                self.set_in_crs(layer.crs)

                # When we have passed bounds:
                if unprojected_bounds:
                    # Set the output CRS, if not yet set, using unprojected bounds.
                    self.set_out_crs(unprojected_bounds)

                    # If we haven't set the projected bounds yet, do that.
                    if not self.projected_bounds:
                        self.update_projected_bounds(self.in_crs, self.out_crs, unprojected_bounds, padding)

                    # Convert global bounds to layer.crs.
                    bounds = bounding.transform(self.out_crs, layer.crs, self.projected_bounds)

                # When we have no passed bounds:
                else:
                    # Set the output CRS, if not yet set, using this layer's bounds.
                    self.set_out_crs(layer.bounds)

                    # Extend projection_bounds
                    self.update_projected_bounds(layer.crs, self.out_crs, layer.bounds, padding)
                    bounds = layer.bounds

                kwargs = self._prepare_layer(layer, path, bounds, **kwargs)
                group = tuple(self.feature(f, **kwargs) for _, f in layer.items(bbox=bounds))

        return {
            'members': group,
            'id': kwargs['name'],
            'class': u' '.join(_style.sanitize(c) for c in layer.schema['properties'].keys())
        }

    def feature(self, feature, transforms, classes, **kwargs):
        '''
        Draw a single feature.

        Args:
            feature (dict): A GeoJSON like feature dict produced by Fiona.
            transforms (list): Functions to apply to the geometry.
            classes (list): Names (unsanitized) of fields to apply as classes in the output element.
            precision (int): rounding precision for coordinates.
            id_field (str): Field to use as id of the output element.
            name (str): layer name (usually basename of the input file).

        Returns:
            str (unicode in Python 2)
        '''
        name = kwargs.pop('name', '?')
        geom = feature.get('geometry')
        precision = kwargs.pop('precision', self.precision)

        try:
            # Check if geometry exists (a bit unpythonic, but cleaner errs this way).
            if geom is None:
                raise ValueError('NULL geometry')

            # Apply transformations to the geometry.
            for t in transforms:
                geom = t(geom) if t is not None else geom

        except ValueError as e:
            self.log.warning('error transforming feature %s of %s: %s',
                             kwargs.get('id', feature.get('id', '?')), name, e)
            return ''

        # Set up the element's properties.
        classes = _style.construct_classes(classes, feature['properties'])

        # Add the layer name to the class list.
        if name != '?':
            classes.insert(0, _style.sanitize(name))

        drawargs = {
            'class': ' '.join(classes)
        }

        if 'id_field' in kwargs:
            drawargs['id'] = _style.sanitize(feature['properties'].get(kwargs['id_field']))

        try:
            # Draw the geometry.
            return draw.geometry(geom, precision=precision, **drawargs)

        except errors.SvgisError as e:
            self.log.warning('unable to draw feature %s of %s: %s',
                             kwargs.get('id', feature.get('id', '?')), name, e)
            return u''

    def compose(self, bounds=None, style=None, viewbox=True, inline=True, **kwargs):
        '''
        Draw files to svg.

        Args:
            bounds (Sequence): Map bounding box in WGS84 (longlat) coordinates.
                               Defaults to map data bounds.
            scalar (int): factor by which to scale the data, generally a small number (1/map scale).
            style (str): CSS to append to parent object CSS.
            viewbox (bool): If True, draw SVG with a viewbox. If False, translate coordinates
                            to the frame. Defaults to True.
            inline (bool): If False, do not add CSS style attributes to each element.
            padding (int): Number of (projected) units to pad bounds by.
            precision (int): Precision for rounding output coordinates.

        Returns:
            String (unicode in Python 2) containing an entire SVG document.
        '''
        # Set up arguments
        scalar = kwargs.pop('scalar', self.scalar)
        bounds = bounding.check(bounds) or self.unprojected_bounds

        # Draw files
        members = [svg.group(**self.compose_file(f, bounds, scalar=scalar, **kwargs))
                   for f in self.files]

        self.log.info('composing drawing')
        drawing = self.draw(members, scalar, kwargs.get('precision'),
                             style=style, viewbox=viewbox, inline=inline)

        # Always reset projected bounds.
        self._projected_bounds = (None,) * 4

        return drawing

    def draw(self, members, scalar=None, precision=None, style=None, **kwargs):
        '''
        Combine drawn layers into an SVG drawing.

        Args:
            members (list): unicode representations of SVG groups.
            scalar (int): factor by which to scale the data, generally a small number (1/map scale).
            style (str): CSS to append to parent object CSS.
            viewbox (bool): If True, draw SVG with a viewbox. If False, translate coordinates to
                            the frame. Defaults to True.
            inline (bool): If True, try to run CSS into each element.

        Returns:
            String (unicode in Python 2) containing an entire SVG document.
        '''
        scalar = scalar or self.scalar
        precision = precision or self.precision
        style = style or self.style
        transform_attrib = 'scale(1,-1)'

        try:
            if any((utils.isinf(b) for b in self._projected_bounds)):
                self.log.warning('Drawing has infinite bounds, consider changing projection or bounding box.')

            dims = [float(b or 0.) * scalar for b in self.projected_bounds]
        except TypeError:
            self.log.warning('Unable to find bounds, the map probably empty ¯\_(ツ)_/¯')
            dims = 0, 0, 0, 0

        # width and height
        size = [dims[2] - dims[0], dims[3] - dims[1]]

        self.log.debug('Size: %f x %f', *size)

        if kwargs.pop('viewbox', True):
            viewbox = [dims[0], -dims[3]] + size
            self.log.debug('drawing with viewbox')
        else:
            viewbox = None
            transform_attrib += ' translate({},{})'.format(-dims[0], -dims[3])
            self.log.debug('translating contents to fit')

        # Create container and then SVG
        container = svg.group(members, fill_rule='evenodd', transform=transform_attrib)
        drawing = svg.drawing(size, [container], style=style, precision=precision, viewbox=viewbox)

        if kwargs.pop('inline', False):
            self.log.info('inlining styles')
            drawing = _style.inline(drawing, style)

        return drawing



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
from svgis import projection
from svgis.utils import DEFAULT_GEOID

SHP = 'tests/test_data/cb_2014_us_nation_20m.json'


class ProjectionTestCase(unittest.TestCase):

    def testUtm(self):
        assert projection.utm_proj4(-21, 42) == '+proj=utm +zone=27 +north +datum=WGS84 +units=m +no_defs'

        assert projection.utm_proj4(-21, -42) == '+proj=utm +zone=27 +south +datum=WGS84 +units=m +no_defs'

        with self.assertRaises(ValueError):
            projection.utm_proj4(-200, 100)

    def testLocalTm(self):
        fixture = ('+proj=lcc +lon_0=0 +lat_1=0 +lat_2=0 +lat_0=0'
                   '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
                   '+units=m +no_defs')
        self.assertEqual(projection.tm_proj4(0, 0, 0), fixture)

    def testGenerateCRS(self):
        bounds = -82.2, 40.1, -78.9, 45.8
        a = projection.generateproj4('utm', bounds=bounds, file_crs=DEFAULT_GEOID)
        self.assertEqual(a, '+proj=utm +zone=17 +north +datum=WGS84 +units=m +no_defs')

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
import re
from xml.dom import minidom
import six
from svgis import svgis, errors


class SvgisTestCase(unittest.TestCase):
    file = 'tests/test_data/cb_2014_us_nation_20m.json'

    polygon = {
        "properties": {
            'apple': 'fruit',
            'pear': 1,
            'kale': 'leafy green',
        },
        "geometry": {
            "type": "Polygon",
            "id": "Polygon",
            "coordinates": [[(0, 0), (1, 0), (0, 1), (0, 0)]]
        }
    }

    def setUp(self):
        self.svgis_obj = svgis.SVGIS(self.file)

    def assertSequenceAlmostEqual(self, a, b):
        for z in zip(a, b):
            self.assertAlmostEqual(*z)

    def testSvgisError(self):
        with self.assertRaises(errors.SvgisError):
            raise errors.SvgisError('This is an error')

    def testSvgisCreate(self):
        self.assertEqual(self.svgis_obj.files, [self.file])
        assert self.svgis_obj._projected_bounds == (None,) * 4
        assert self.svgis_obj.out_crs is None
        assert self.svgis_obj.style == svgis.STYLE

        svgis_obj2 = svgis.SVGIS([self.file])
        assert svgis_obj2.files == [self.file]

        with self.assertRaises(ValueError):
            svgis.SVGIS(12)

    def testSvgisCompose(self):
        composed = self.svgis_obj.compose()
        assert isinstance(composed, six.string_types)

    def testSvgisClassFields(self):
        composed = self.svgis_obj.compose(class_fields=('NAME', 'GEOID'))

        matchiter = re.finditer(r'class="(.+?)"', composed)
        match = next(matchiter)

        self.assertIsNotNone(match)
        self.assertIn(u'AFFGEOID', match.groups()[0])
        self.assertIn(u'GEOID', match.groups()[0])
        self.assertIn(u'NAME', match.groups()[0])

        match = next(matchiter)
        self.assertIn('GEOID_US', match.groups()[0])
        self.assertIn('cb_2014_us_nation_20m', match.groups()[0])

    def testRepr(self):
        expected = ("SVGIS(files=['{}'], "
                    'out_crs=None)'.format(self.file))
        self.assertEqual(str(self.svgis_obj), expected)

    def testDrawGeometry(self):
        feat = {
            "geometry": {
                'type': 'LineString',
                'coordinates': [[-110.8, 35.3], [-110.9, 35.8], [-110.5, 35.1], [-110.8, 35.3]]
            },
            "properties": {
                'foo': 'bar',
                'cat': 'meow'
            }
        }
        drawn = self.svgis_obj.feature(feat, [], classes=['foo'], id_field='cat', name='quux')
        assert isinstance(drawn, six.string_types)

        self.assertIn('id="meow"', drawn)
        self.assertIn('class="quux foo_bar"', drawn)

        drawn2 = self.svgis_obj.feature(feat, [], classes=['foo'], id_field='cat')
        self.assertIn('class="foo_bar"', drawn2)

        feat['geometry'] = None
        drawn3 = self.svgis_obj.feature(feat, [], classes=['foo'], id_field='cat')
        assert drawn3 == ''

    def testSvgisComposeType(self):
        a = self.svgis_obj.compose(inline_css=True)
        b = self.svgis_obj.compose(inline_css=False)

        try:
            try:
                self.assertIsInstance(a, unicode)
            except AssertionError:
                raise AssertionError(type(a))

            try:
                self.assertIsInstance(b, unicode)
            except AssertionError:
                raise AssertionError(type(b))

        except NameError:
            try:
                self.assertIsInstance(a, str)
            except AssertionError:
                raise AssertionError(type(a))

            try:
                self.assertIsInstance(b, str)
            except AssertionError:
                raise AssertionError(type(b))

        self.assertEqual(type(a), type(b))

    def testMapFunc(self):
        args = {
            "scale": 1000,
            "padding": 10,
            "inline_css": True,
            "clip": False,
            'crs': 'EPSG:32117',
        }
        result = svgis.map([self.file], (-80, 40, -71, 45.1), **args)

        self.assertIn(svgis.STYLE, result)

        doc = minidom.parseString(result)
        for poly in doc.getElementsByTagName('polygon'):
            self.assertIn('polygon', poly.toxml())
            self.assertIn('style', poly.toxml())
            style = poly.getAttribute('style')
            self.assertIn('fill:none', style)
            self.assertIn('stroke-linejoin:round', style)

    def testOpenZips(self):
        archive = 'zip://tests/test_data/test.zip/tests/test_data/cb_2014_us_nation_20m.json'

        result = svgis.SVGIS([archive], simplify=60).compose()

        self.assertIn('cb_2014_us_nation_20m', result)

    def testDrawWithClasses(self):
        r0 = self.svgis_obj.feature(self.polygon, [], classes=[], id_field=None, name='potato')
        self.assertIn('class="potato"', r0)

        r1 = self.svgis_obj.feature(self.polygon, [], classes=['kale'], id_field='apple')
        assert 'kale_leafy_green' in r1
        assert 'id="fruit"' in r1

        r2 = self.svgis_obj.feature(self.polygon, [], id_field='pear', classes=['apple', 'pear', 'kale'])
        self.assertIn('id="_1"', r2)
        assert 'apple_fruit' in r2
        assert 'pear_1' in r2
        assert 'kale_leafy_green' in r2

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Draw geometries as SVG elements'''

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

from functools import wraps
from fionautil import measure
from . import svg, transform
from .errors import SvgisError


def _applyid(multifunc):
    '''
    This decorator applies the ID attribute to the group that
    contains multi-part geometries, rather than the elements of the group.
    '''
    @wraps(multifunc)
    def func(coordinates, **kwargs):
        ID = kwargs.pop('id', None)
        result = svg.group(multifunc(coordinates, **kwargs), fill_rule="evenodd", id=ID)
        return result

    return func


def linestring(coordinates, **kwargs):
    return svg.polyline(coordinates, **kwargs)


@_applyid
def multilinestring(coordinates, **kwargs):
    return (linestring(coords, **kwargs) for coords in coordinates)


def lines(geom, **kwargs):
    '''
    Draw a LineString or MultiLineString geometry.

    Args:
        geom (object): A GeoJSON-like LineString or MultiLineString geometry object.

    Returns:
        str (unicode in Python 2) representation of the SVG group or polyline element(s).
    '''
    if geom['type'] == 'LineString':
        return linestring(geom['coordinates'], **kwargs)

    elif geom['type'] == 'MultiLineString':
        return multilinestring(geom['coordinates'], **kwargs)


def polygons(geom, **kwargs):
    '''
    Draw polygon(s) in a feature. transform is a function to operate on coords.
    Draws first ring clockwise, and subsequent ones counter-clockwise.

    Args:
        geom (object): A GeoJSON-like Polygon or MultiPolygon geometry object.

    Returns:
        str (unicode in Python 2) representation of the SVG group, path or polygon element.
    '''
    if geom['type'] == 'Polygon':
        return polygon(geom['coordinates'], **kwargs)

    elif geom['type'] == 'MultiPolygon':
        return multipolygon(geom['coordinates'], **kwargs)


def polygon(coordinates, **kwargs):
    if len(coordinates) == 1:
        return svg.polygon(coordinates[0], **kwargs)

    kwargs['class'] = ('polygon ' + kwargs.pop('class', '')).strip()

    instructions = ['M']
    # This is trickier because drawing holes in SVG.
    # We go clockwise on the first ring, then counterclockwise
    if measure.counterclockwise(coordinates[0]):
        instructions.extend(coordinates[0][::-1])
    else:
        instructions.extend(coordinates[0])

    instructions.append('z')

    for ring in coordinates[1:]:
        # make all interior run the counter-clockwise
        instructions.append('M')
        instructions.extend(ring[::-1] if measure.clockwise(ring) else ring)
        instructions.append('z')

    return svg.path(instructions, fill_rule='evenodd', **kwargs)


@_applyid
def multipolygon(coordinates, **kwargs):
    return (polygon(coords, **kwargs) for coords in coordinates)


def points(geom, **kwargs):
    '''
    Draw a Point or MultiPoint geometry

    Args:
        geom (object): A GeoJSON-like Point or MultiPoint geometry object.

    Returns:
        str (unicode in Python 2) representation of the SVG group, or circle element
    '''
    kwargs['r'] = kwargs.get('r', 1)

    if geom['type'] == 'Point':
        return svg.circle(geom['coordinates'], **kwargs)

    elif geom['type'] == 'MultiPoint':
        return multipoint(geom['coordinates'], **kwargs)


@_applyid
def multipoint(coordinates, **kwargs):
    return (svg.circle((pt[0], pt[1]), **kwargs) for pt in coordinates)


def geometrycollection(collection, bbox, precision, **kwargs):
    ID = kwargs.pop('id', None)
    geoms = (geometry(g, bbox=bbox, precision=precision, **kwargs)
             for g in collection['geometries'])
    return svg.group(geoms, fill_rule="evenodd", id=ID)


def geometry(geom, bbox=None, precision=None, **kwargs):
    '''
    Draw a geometry. Will return either a single geometry or a group.

    Args:
        geom (object): A GeoJSON-like geometry object. Coordinates must be 2-dimensional.
        bbox (tuple): An optional bounding minimum bounding box
        precision (int): Rounding precision, must be 0 or greater (default: no rounding).
        kwargs (object): keyword args to be passed onto the created
                         elements (e.g. class, id, style).

    Returns:
        str (unicode in Python 2) representation of SVG element(s) of the given geometry.
    '''
    if bbox:
        geom = transform.clip(geom, bbox)

    if geom['type'] in ('Point', 'MultiPoint'):
        return points(geom, precision=precision, **kwargs)

    elif geom['type'] in ('LineString', 'MultiLineString'):
        return lines(geom, precision=precision, **kwargs)

    elif geom['type'] in ('Polygon', 'MultiPolygon'):
        return polygons(geom, precision=precision, **kwargs)

    elif geom['type'] == 'GeometryCollection':
        return geometrycollection(geom, bbox, precision, **kwargs)

    else:
        raise SvgisError("Can't draw features of type: {}".format(geom['type']))


def group(geometries, **kwargs):
    '''
    Add a list of geometries to a group.

    Args:
        geometries (Sequence): GeoJSON-like geometry dicts.

    Returns:
        str (unicode in Python 2) representation of the SVG group
    '''
    return svg.group([geometries(g, fill_rule="evenodd", **kwargs) for g in geometries])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Utilities for svgis'''

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
from math import ceil, floor
from itertools import groupby


# WGS 84
DEFAULT_GEOID = {'init': 'EPSG:4326', 'no_defs': True, 'proj': 'longlat'}


def posint(i):
    '''Fake class for positive integers only'''
    I = int(i)
    if I < 0:
        raise ValueError('Must be a positive integer')
    return I


def isinf(x):
    inf = float('inf')
    return x == inf or x == inf * -1


def between(a, b, count=None):
    """Yield <count> points between two floats"""
    jump = (b - a) / float(count or 10.)

    while a < b:
        yield a
        a += jump


def frange(a, b, step, cover=None):
    """like range, but for floats."""
    while a < b:
        yield a
        a += step

    if cover:
        yield a


def modfloor(inp, mod):
    i = inp - (inp % mod)
    return int(floor(i))


def modceil(inp, mod):
    i = inp + mod - (inp % mod)
    return int(ceil(i))


def rnd(i, precision):
    if precision is None:
        return i
    else:
        return round(i, precision)


def dedupe(array):
    '''
    Use itertools.groupby to remove duplicates in a list.
    '''
    try:
        array = array.tolist()
    except AttributeError:
        pass

    for g in groupby(array):
        yield g[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import sys
from signal import signal, SIGPIPE, SIG_DFL
import logging
import click
import fiona.crs
from fionautil.layer import meta_complete
from . import bounding, projection
from . import graticule as _graticule, style as _style, svgis, __version__
from .utils import DEFAULT_GEOID, posint

none = {
    'flag_value': None,
    'expose_value': False,
    'help': '(not enabled)'
}

try:
    import shapely
    clipkwargs = {
        'default': True,
        'flag_value': True,
        'help': "Clip shapes to bounds. Slightly slower, produces smaller files (default: clip)."
    }
except ImportError:
    clipkwargs = none

try:
    import visvalingamwyatt
    simplifykwargs = {
        'type': click.IntRange(1, 100, clamp=True),
        'metavar': 'FACTOR',
        'help': ('Simplify geometries, '
                 'accepts an integer between 1 and 100, '
                 'the percentage of each geometry to retain.'),
    }
except ImportError:
    simplifykwargs = none

CLICKARGS = {
    'context_settings': dict(help_option_names=['-h', '--help'])
}

csskwargs = {
    'flag_value': True,
    'default': True,
    'help': ('Inline CSS styles to each element. '
             'Slightly slower, but required by some clients (e.g. Adobe) '
             '(default: inline).'),
}

inp = click.argument('layer', default=sys.stdin, type=click.File('rb'))
outp = click.argument('output', default=sys.stdout, type=click.File('wb'))


# Base
@click.group(**CLICKARGS)
@click.version_option(version=__version__, message='%(prog)s %(version)s')
@click.pass_context
def main(context):
    context.log = logging.getLogger('svgis')
    context.log.setLevel(logging.WARN)
    ch = logging.StreamHandler()
    ch.setLevel(logging.WARN)
    context.log.addHandler(ch)

    signal(SIGPIPE, SIG_DFL)


# Style
style_help = ("Style to append to SVG. "
              "Either a valid CSS string, a file path (must end in '.css'). "
              "Use '-' for stdin.")


@main.command()
@inp
@outp
@click.option('-c', '--style', type=str, help=style_help, default='')
@click.option('-r', '--replace', flag_value=True, help='Replace existing styles')
@click.option('--inline/--no-inline', '-l/ ', **csskwargs)
def style(layer, output, **kwargs):
    """Add or inline the CSS styles of an SVG"""
    result = _style.add_style(layer, kwargs['style'], kwargs['replace'])
    if kwargs['inline']:
        result = _style.inline(result)
    click.echo(result.encode('utf-8'), file=output)


@main.command()
@inp
@outp
@click.option('-f', '--scale', type=int)
def scale(layer, output, **kwargs):
    '''Scale all coordinates in an SVG by a factor'''
    click.echo(_style.rescale(layer, factor=kwargs['scale']).encode('utf-8'), file=output)


crs_help = ('Specify a map projection. '
            'Accepts either an EPSG code (e.g. epsg:4456), '
            'a proj4 string, '
            'a file containing a proj4 string, '
            '"utm" (use local UTM), '
            '"file" (use existing), '
            '"local" (generate a local projection)')


@main.command()
@click.argument('layer', default=sys.stdin, type=click.Path(exists=True))
@click.option('-j', '--crs', type=str, metavar='KEYWORD', default='file', help=crs_help + ' (default: file)')
@click.option('--latlon', default=False, flag_value=True, help='Print bounds in latitude, longitude order')
def bounds(layer, crs, latlon=False):
    '''Return the bounds for a given layer, optionally projected.'''
    meta = meta_complete(layer)

    # If crs==file, these will basically be no ops.
    out_crs = projection.pick(crs, meta['bounds'], file_crs=meta['crs'])
    result = bounding.transform(meta['crs'], out_crs, meta['bounds'])

    if latlon:
        fmt = '{0[1]} {0[0]} {0[3]} {0[2]}'
    else:
        fmt = '{0[0]} {0[1]} {0[2]} {0[3]}'

    click.echo(fmt.format(result), file=sys.stdout)


# Draw
@main.command()
@click.argument('layer', nargs=-1, type=str, required=True)
@click.option('-o', '--output', default=sys.stdout, type=click.File('wb'), help="Defaults to stdout")
@click.option('-b', '--bounds', nargs=4, type=float, metavar="minx miny maxx maxy",
              help='In the same coordinate system as the first input layer', default=None)
@click.option('-c', '--style', type=str, metavar='CSS', help="CSS file or string", multiple=True)
@click.option('-f', '--scale', type=int, default=None,
              help='Scale for the map (units are divided by this number)')
@click.option('-p', '--padding', type=int, default=None, required=None,
              help='Buffer the map (in projection units)')
@click.option('-i', '--id-field', type=str, metavar='FIELD', help='Geodata field to use as ID')
@click.option('-a', '--class-fields', type=str, default='', metavar='FIELDS', multiple=True,
              help='Geodata fields to use as class (comma-separated)')
@click.option('-j', '--crs', metavar='KEYWORD', type=str, help=crs_help)
@click.option('-s', '--simplify', **simplifykwargs)
@click.option('-P', '--precision', metavar='INTEGER', type=posint, default=5,
              help='Rounding precision for coordinates (default: 5)')
@click.option('--clip/--no-clip', ' /-n', **clipkwargs)
@click.option('--inline/--no-inline', '-l/ ', **csskwargs)
@click.option('--viewbox/--no-viewbox', ' /-x', default=False, help='Draw SVG using a ViewBox (default: no ViewBox)')
@click.option('-q', '--quiet', default=False, flag_value=True, help='Ignore warnings')
@click.option('-v', '--verbose', default=False, count=True, help='Talk a lot')
def draw(layer, output, **kwargs):
    '''Draw SVGs from input geodata'''
    log = logging.getLogger('svgis')

    verbose = kwargs.pop('verbose', None)
    if verbose:
        level = logging.DEBUG if verbose > 1 else logging.INFO
        log.setLevel(level)
        for h in log.handlers:
            h.setLevel(level)

    if kwargs.pop('quiet', None):
        log.handlers[0].setLevel(logging.ERROR)
        log.setLevel(logging.ERROR)

    click.echo(svgis.map(layer, **kwargs).encode('utf-8'), file=output)
    log.info('writing %s', output.name)


# Proj
@main.command()
@click.argument('bounds', nargs=4, type=float, metavar="minx miny maxx maxy", default=None)
@click.option('-m', '--method', default='local', type=click.Choice(('utm', 'local')), help='Defaults to local')
@click.option('-j', '--crs', default=DEFAULT_GEOID, help='Projection of the bounding coordinates')
def project(bounds, method, crs):
    '''Get a local Transverse Mercator or UTM projection for a bounding box. Expects WGS84 coordinates.'''
    if crs in projection.METHODS:
        click.echo('CRS must be an EPSG code, a Proj4 string, or file containing a Proj4 string.', err=1)
    result = fiona.crs.to_string(projection.pick(method, file_crs=crs, bounds=bounds))
    click.echo(result.encode('utf-8'))


crs_help2 = ('Specify a map projection. '
             'Accepts either an EPSG code (e.g. epsg:4456), '
             'a proj4 string, '
             'a file containing a proj4 string, '
             '"utm" (use local UTM), '
             '"local" (generate a local projection)')


# Graticule
@main.command()
@click.argument('bounds', nargs=4, type=float, metavar='minx miny maxx maxy')
@click.option('-s', '--step', type=float, help='Step between lines (in projected units)', required=True)
@click.option('-j', '--crs', type=str, default=None, help=crs_help2)
@click.option('-o', '--output', default=sys.stdout, type=click.File('wb'), help="Defaults to stdout")
def graticule(bounds, step, crs, output):
    '''Generate a GeoJSON containing a graticule. Accepts a bounding box in longitude and latitude (WGS84).'''
    click.echo(_graticule.geojson(bounds, step, crs), file=output)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Deal with projections'''

# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>
import os.path
import fiona.transform
import fiona.crs
import pyproj
import utm
from six import string_types
from .utils import DEFAULT_GEOID
from . import bounding


METHODS = 'default', 'file', 'local', 'utm'


def tm_proj4(x0, y0, y1):
    '''
    Generate the proj4 string for a local Transverse Mercator projection
    centered at a given longitude and between two latitudes.

    Args:
        x0 (float): longitude
        y0 (float): latitude 0
        y1 (float): latitude 1

    Returns:
        (str) proj4 string
    '''
    return ('+proj=lcc +lon_0={x0} +lat_1={y1} +lat_2={y0} +lat_0={y1}'
            '+x_0=0 +y_0=0 +ellps=GRS80 +towgs84=0,0,0,0,0,0,0'
            '+units=m +no_defs').format(x0=x0, y0=y0, y1=y1)


def utm_proj4(lon, lat):
    '''
    Generate the proj4 string for the UTM projection at a given (lon, lat) coordinate.

    Args:
        lon (float): longitude
        lat (float): latitude

    Returns:
        (str) proj4 string
    '''
    try:
        _, _, zonenumber, zoneletter = utm.from_latlon(lat, lon)

        if zoneletter in 'ZYXWVUTSRQPN':
            hemisphere = 'north'

        elif zoneletter in 'MLKJHGFEDCBA':
            hemisphere = 'south'

        return '+proj=utm +zone={} +{} +datum=WGS84 +units=m +no_defs'.format(zonenumber, hemisphere)

    except utm.error.OutOfRangeError as e:
        raise ValueError(e)


def generateproj4(method, bounds, file_crs):
    '''
    Generate a Proj4 projection definition: either the local UTM zone
    or a custom transverse mercator.

    Args:
        method (str): If default, local or utm.
            * If 'utm': generate a UTM. Otherwise, a local projection,
            * If 'local': generate a custom local transverse mercator projection,
            * If 'default': if file_crs is longlat, act like local. otherwise use file_crs
        bounds (tuple): bounding box
        file_crs (dict): Fiona-generated CRS of the input file

    Returns:
        (str) proj4 string
    '''
    if bounds is None or file_crs is None:
        raise ValueError('generatecrs missing bounds and file crs')

    is_longlat = _is_longlat(file_crs)

    if method == 'default':
        # Check if file_crs _is_longlat, if not use that.
        # Option to continue returning default if we didn't get a file_crs
        if is_longlat:
            method = 'local'
        else:
            return fiona.crs.to_string(file_crs)

    if is_longlat:
        longlat_bounds = bounds
    else:
        longlat_bounds = bounding.transform(file_crs, DEFAULT_GEOID, bounds)

    minx, miny, maxx, maxy = longlat_bounds

    if method == 'utm':
        midx = (minx + maxx) / 2
        midy = (miny + maxy) / 2

        return utm_proj4(midx, midy)

    elif method == 'local':
        # Create a custom TM projection
        x0 = (float(minx) + float(maxx)) // 2

        return tm_proj4(x0, miny, maxy)


def _is_longlat(crs):
    '''Test if CRS is in lat/long coordinates'''
    try:
        if crs.get('proj') == 'longlat' or pyproj.Proj(**crs).is_latlong():
            return True
    except RuntimeError:
        pass

    return False


def pick(project, bounds=None, file_crs=None):
    '''
    Pick a projection or projection method to use.

    Returns:
        (mixed) one of: None, 'local', 'utm' or a dict
    '''
    out_crs = None
    project = project or 'default'

    if isinstance(project, dict):
        return project

    elif isinstance(project, string_types):
        if project.lower() == 'file':
            out_crs = file_crs if file_crs is not None else 'file'

        # Is an epsg code
        elif project.lower()[:5] == 'epsg:':
            out_crs = fiona.crs.from_epsg(int(project.split(':')[1]))

        elif project.lower() in METHODS:
            try:
                out_crs = fiona.crs.from_string(generateproj4(project, bounds, file_crs))

            except ValueError:
                return project

        # Is a file
        elif os.path.exists(project):
            with open(project) as f:
                out_crs = fiona.crs.from_string(f.read())

        else:
            # Assume it's a proj4 string.
            # fiona.crs.from_string returns {} if it isn't.
            out_crs = fiona.crs.from_string(project)

    return out_crs


def fake_to_string(crs):
    '''
    Fake to_string for debugging in places where fiona.crs.to_string
    isn't available
    '''
    return ' '.join('+{0[0]}={0[1]}'.format(i) for i in crs.items())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import unittest
import re
import six
from svgis import draw, errors, svgis


class DrawTestCase(unittest.TestCase):

    properties = {
        'cat': u'meow',
        'dog': 'woof'
    }

    classes = [u'foo', 'cat']

    lis1 = [[-110.6, 35.3], [-110.7, 35.5], [-110.3, 35.5], [-110.2, 35.1], [-110.2, 35.8], [-110.3, 35.2],
            [-110.1, 35.8], [-110.8, 35.5], [-110.7, 35.7], [-110.1, 35.4], [-110.7, 35.1], [-110.6, 35.3]]

    lis2 = [[-110.8, 35.3], [-110.6, 35.4], [-110.1, 35.5], [-110.1, 35.5], [-110.4, 35.2],
            [-110.5, 35.1], [-110.5, 35.1], [-110.9, 35.8], [-110.5, 35.1], [-110.8, 35.3]]

    def setUp(self):
        self.multipolygon = {
            "properties": self.properties,
            "geometry": {
                "type": "MultiPolygon",
                "id": "MultiPolygon",
                "coordinates": [[self.lis1], [self.lis2]]
            }
        }
        self.polygon = {
            "properties": self.properties,
            "geometry": {
                "type": "Polygon",
                "id": "Polygon",
                "coordinates": [self.lis1]
            }
        }
        self.multilinestring = {
            "properties": self.properties,
            "geometry": {
                'type': 'MultiLineString',
                "id": "MultiLineString",
                'coordinates': [self.lis2, self.lis2]
            }
        }
        self.linestring = {
            "properties": self.properties,
            "geometry": {
                'coordinates': self.lis2,
                'type': 'LineString',
                "id": "LineString",
            }
        }
        self.point = {
            "properties": self.properties,
            "geometry": {
                'coordinates': (0.0, 0),
                'type': 'Point',
                "id": "Point",
            }
        }

        self.obj = svgis.SVGIS([])

    def testDrawPoint(self):
        feat = self.obj.feature(self.point, [], classes=self.classes, id_field=None)

        assert isinstance(feat, six.string_types)
        self.assertIn('cat_meow', feat)

    def testDrawLine(self):
        line = draw.lines(self.linestring['geometry'])
        assert isinstance(line, six.string_types)

        feat = self.obj.feature(self.linestring, [], classes=self.classes, id_field=None)

        assert isinstance(feat, six.string_types)
        assert 'cat_meow' in feat

    def testDrawMultiLine(self):
        mls1 = draw.multilinestring(self.multilinestring['geometry']['coordinates'])
        mls2 = draw.lines(self.multilinestring['geometry'])

        assert isinstance(mls1, six.string_types)
        assert isinstance(mls2, six.string_types)

        grp = self.obj.feature(self.multilinestring, [], classes=self.classes, id_field=None)

        assert isinstance(grp, six.string_types)
        assert 'cat_meow' in grp

    def testDrawPolygon(self):
        drawn = draw.polygon(self.polygon['geometry']['coordinates'])
        assert "{},{}".format(*self.lis1[0]) in drawn
        feat = self.obj.feature(self.polygon, [], classes=self.classes, id_field=None)
        assert 'cat_meow' in feat

    def testDrawMultiPolygon(self):
        drawn = draw.multipolygon(self.multipolygon['geometry']['coordinates'])

        assert isinstance(drawn, six.string_types)

    def testDrawMultiPoint(self):
        points = draw.multipoint(self.lis1, id='foo')

        self.assertIn('cy="35.1"', points)
        self.assertIn('cx="-110.6"', points)
        assert re.search(r'<g[^>]*id="foo"', points)

    def testAddClass(self):
        geom = {
            'coordinates': (0, 0),
            'type': 'Point'
        }
        kwargs = {
            "class": "boston"
        }
        point = draw.points(geom, **kwargs)
        self.assertIsInstance(point, six.string_types)

        point = draw.points(geom, **kwargs)
        assert isinstance(point, six.string_types)

    def testDrawPolygonComplicated(self):
        coordinates = [
            [(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0), (0.0, 0.0)],
            [(4.0, 4.0), (4.0, 5.0), (5.0, 5.0), (5.0, 4.0), (4.0, 4.0)]
        ]

        polygon = draw.polygon(coordinates)
        self.assertIsInstance(polygon, six.string_types)
        assert 'class="polygon"' in polygon

        kw = {'class': 'a'}
        assert 'polygon a' in draw.polygon(coordinates, **kw)

    def testUnkownGeometry(self):
        with self.assertRaises(errors.SvgisError):
            draw.geometry({"type": "FooBar", "coordinates": []})

    def testGeometryCollection(self):
        gc = {
            "type": "GeometryCollection",
            "id": "GC",
            "geometries": [
                self.polygon['geometry'],
                self.linestring['geometry'],
                self.point['geometry'],
                self.multipolygon['geometry'],
                self.multilinestring['geometry']
            ],
        }
        a = draw.geometry(gc, id='cats')
        assert isinstance(a, six.string_types)
        assert 'id="cats"' in a

    def testDrawAndConvertToString(self):
        draw.geometry(self.linestring['geometry'])
        draw.geometry(self.multilinestring['geometry'])
        draw.geometry(self.polygon['geometry'])
        draw.geometry(self.multipolygon['geometry'])

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Clip and simplify geometries'''

# This file is part of svgis.
# https://github.com/fitnr/svgis
# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>
from types import GeneratorType
from functools import partial
try:
    from shapely.geometry import shape, mapping
    from shapely.geos import TopologicalError
    import numpy as np
except ImportError:
    pass
try:
    import visvalingamwyatt as vw
except ImportError:
    pass


def _expand_np(coordinates):
    return np.array(_expand_py(coordinates))


def _expand_py(coordinates):
    return tuple(coordinates)


def expand(ring):
    if isinstance(ring, GeneratorType):
        try:
            return _expand_np(ring)
        except NameError:
            return _expand_py(ring)
    else:
        return ring


def expand_rings(rings):
    try:
        return tuple(_expand_np(ring) for ring in rings)
    except NameError:
        return tuple(_expand_py(ring) for ring in rings)


def expand_geom(geom):
    '''Expand generators in a geometry's coordinates.'''

    if geom['type'] == 'MultiPolygon':
        geom['coordinates'] = tuple(expand_rings(rings) for rings in geom['coordinates'])

    elif geom['type'] in ('Polygon', 'MultiLineString'):
        geom['coordinates'] = expand_rings(geom['coordinates'])

    elif geom['type'] in ('Point', 'MultiPoint', 'LineString'):
        geom['coordinates'] = expand(geom['coordinates'])

    elif geom['type'] == 'GeometryCollection':
        geom['geometries'] = tuple(expand_geom(g) for g in geom['geometries'])

    else:
        raise NotImplementedError("Unsupported geometry type " + geom['type'])

    return geom


def clipper(bbox):
    '''
    Create a clipping function for a given bounding box.

    Args:
        bbox (tuple): bounding box

    Returns:
        function that will given geometries to input bounding box
    '''
    minx, miny, maxx, maxy = bbox
    bounds = {
        "type": "Polygon",
        "coordinates": [[(minx, miny), (minx, maxy), (maxx, maxy), (maxx, miny), (minx, miny)]]
    }
    try:
        bbox_shape = shape(bounds)

        def func(geometry):
            # This is technically only needed in Py3, but whatever.
            geom = expand_geom(geometry)

            try:
                clipped = bbox_shape.intersection(shape(geom))
            except (ValueError, TopologicalError):
                return geometry

            return mapping(clipped)

    except NameError:
        def func(geometry):
            return geometry

    return func


def clip(geometry, bounds):
    '''
    Clip a geometry to a bounding box. Equivalent to calling clipper(bounds)(geometry).

    Args:
        geometry (dict): geometry object
        bounds (tuple): bounding box

    Returns:
        (dict) geometry
    '''
    try:
        return clipper(bounds)(geometry)

    except NameError:
        return geometry


def simplifier(ratio):
    '''
    Create a simplification function, if visvalingamwyatt is available.
    Otherwise, return a noop function.

    Args:
        ratio (int): Between 1 and 99

    Returns:
        simplification function
    '''
    try:
        # put this first to get NameError out of the way
        simplify = vw.simplify_geometry

        if ratio is None or ratio >= 100 or ratio < 1:
            raise ValueError

        return partial(simplify, ratio=ratio / 100.)

    except (TypeError, ValueError, NameError):
        return None
